/*
 * Copyright 1996 by Michel D. Grimminck
 *
 * All Rights Reserved
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <stdio.h>
#include "const.h"
#include "var.h"
#include "functions.h"
#include <time.h>
#ifndef MAPPEDMEMORY
#ifdef USE_ZLIB
#include "/usr/include/zlib.h"
#endif
#endif

#define DB_INDEX_FILE "endgame.ini"
#define DB_UNKNOWN 3
#define DB_ALL 100
#define CAPTURE 100
#define PROMOTE 101
#define DB_VERIFY 200
#define QMAX 70000000

/* XXvXX log: pentium 100
db:0202, found=139512, time=80
db:0202, found=10415, time=245
db:0202, found=115, time=238
db:0202, found=0, time=240

/* XXvXX log: amd 1667
db:0202, found=139512, time=3
db:0202, found=10415, time=9
db:0202, found=115, time=1
db:0202, found=0, time=0
2v2 tot OOvOO: 160 sec

Generation time for 3v2 + 4v1=4.8 hour

*/

FILE *logfile;
int pos_count[4096];
DBINDEX bytesize[4096];
int verified[4096];
int debugMode=false;
INT64 qsize[2],qmax[2]; /* qsize must be signed */
char qname[2][20]={"/tmp/pos_queue0","/tmp/pos_queue1"};
FILE *qfile[2]={NULL,NULL};
DBINDEX allocatedMemory;

/* 'database' points to all databases with a maximum of 8 man and 8 crowns
   or contains NULL if the database is not present.

   All databases are computed for 'white to move'. If black is to move
   the board is reversed.

*/


int databaseIsLoaded(int wman,int wcrown,int bman,int bcrown)
/* returns true is database is available in memory */
{
    int nr;
    
    nr=database_nr(0,wman,wcrown,bman,bcrown);
    #ifdef MAPPEDMEMORY
        if (mem64db[nr]>=0) return true;
    #else
        if (database[nr]!=NULL) return true;
    #endif
    return false;
}

void init_queue_write(int q)
{
    if (q!=0 && q!=1) {printf("fatal: queue does not exist\n"); exit(1);}
    qsize[q]=0;
    qfile[q]=fopen(qname[q],"wb");
    if (qfile[q]==NULL) {printf("warning: can not open queue for writing\n"); return;}
}

void init_queue_read(int q)
{
    if (q!=0 && q!=1) {printf("fatal: queue does not exist\n"); exit(1);}
    if (qsize[q]<=0) {printf("fatal: attempting to open incomplete queue\n"); exit(1);}
    qfile[q]=fopen(qname[q],"rb");
    if (qfile[q]==NULL) {printf("fatal: can not open queue for reading\n"); exit(1);}
}

void add_to_queue(int q)
{
    int i,n;

    if (qfile[q]==NULL) return; /* queue not open (anymore) */
    if (qsize[q]>=qmax[q]) {
        printf("warning: maximal queue size reached: closing queue\n");
        fclose(qfile[q]);
        qfile[q]=NULL;
        qsize[q]=-1; /* queue incomplete: don't use it */
        return;
    }
    n=0; for(i=0;i<50;i++) if (board[map[i]]!=empty) n++;
    fputc(n,qfile[q]);
    for(i=0;i<50;i++) if (board[map[i]]!=empty) {
            fputc(i,qfile[q]);
            fputc(board[map[i]],qfile[q]);
        }
    qsize[q]++;
}

static char* read_from_queue(int q)
{
    int i,n,p;
    static char local[93];

    if (qfile[q]==NULL) {printf("fatal: attempting to read from closed queue\n"); exit(1);}
    if (feof(qfile[q])) {
        fclose(qfile[q]);
        qfile[q]=NULL;
        return(false);
    }

    for(i=0;i<50;i++) local[map[i]]=empty;
    n=fgetc(qfile[q]);
    for(i=0;i<n;i++) {
        p=fgetc(qfile[q]);
        local[map[p]]=fgetc(qfile[q]);
    }
    return(local);
}

void close_queue(int q)
{
    if (qfile[q]!=NULL) fclose(qfile[q]);
    qfile[q]=NULL;
}
char *database_name(int wman,int wcrown,int bman,int bcrown)
{
    static char name[40];
    int a=0,i;

    for(i=0;i<wcrown;i++) name[a++]='X';
    for(i=0;i<wman;i++) name[a++]='O';
    name[a++]='v';
    for(i=0;i<bcrown;i++) name[a++]='X';
    for(i=0;i<bman;i++) name[a++]='O';
    name[a]='\0';
    return(name);
}

void database_piecelist(int *piecelist,int wman,int wcrown,int bman,int bcrown)
{
    int a=0,i;

    for(i=0;i<wcrown;i++) piecelist[a++]=white|crown;
    for(i=0;i<wman;i++) piecelist[a++]=white|man;
    for(i=0;i<bcrown;i++) piecelist[a++]=black|crown;
    for(i=0;i<bman;i++) piecelist[a++]=black|man;
}

void database_count(char *id,int *wman,int *wcrown,int *bman,int *bcrown)
/* returns the number of pieces for a database */
{
    int n;

    *wman=*bman=*wcrown=*bcrown=0;

    printf("s:%s\n",id);
    n=0;
    while(id[n]!='v') {
        if (id[n]=='X') *wcrown++;
        if (id[n]=='O') *wman++;
        n++;
    }
    while(id[n]!=0) {
        if (id[n]=='X') *bcrown++;
        if (id[n]=='O') *bman++;
        n++;
    }
    printf("s:%i %i\n",n,*wman);
}

int load_database(int wman,int wcrown,int bman,int bcrown)
{
    char db_file[100],*id;
    DBINDEX i;
    int nr;
    DBINDEX size;
    FILE *db;
    int use_compression=false;
    char *p;

#ifndef MAPPEDMEMORY    
#ifdef USE_ZLIB
    gzFile in;
    int len;
#endif
#endif
    id=database_name(wman,wcrown,bman,bcrown);
#ifdef USE_ZLIB
    sprintf(db_file,"databases/%s.raw.gz",id);
#else
    sprintf(db_file,"databases/%s.raw",id);
#endif
    size=db_count(wman,wcrown,bman,bcrown);
    nr=database_nr(white,wman,wcrown,bman,bcrown);
    pos_count[nr]=size;
    bytesize[nr]=(size+3)/4;
    if (!availableOnDisk(wman,wcrown,bman,bcrown)) {
        printf("database %s not found\n",db_file);
        #ifdef MAPPEDMEMORY
            mem64db[nr]=-1;
        #else
            database[nr]=NULL;
        #endif
        return(false);
    }
        
    #ifdef MAPPEDMEMORY
        mem64db[nr]=mem64_allocate(bytesize[nr]);
        if (mem64db[nr]<0) {
            printf("fatal: no memory for database (%i Mb)\n",bytesize[nr]/1024/1024);
            exit(1);
        }
    #else
        database[nr]=(unsigned char *)malloc(sizeof(unsigned char)*bytesize[nr]);
        if (database[nr]==NULL) {
            printf("fatal: no memory for database (%i Mb)\n",bytesize[nr]/1024/1024);
            exit(1);
        }
    #endif
    allocatedMemory+=sizeof(unsigned char)*bytesize[nr];
/*
#ifndef MAPPEDMEMORY
#ifdef USE_ZLIB
    sprintf(db_file,"databases/%s.raw.gz",id);
    db=my_fopen(db_file,"rb");
    if (db!=NULL) use_compression=true;
#endif
#else
    if (use_compression==false) {
        sprintf(db_file,"databases/%s.raw",id);
        db=my_fopen(db_file,"rb");
        if (db==NULL) {
            printf("database %s not found\n",db_file);
            #ifdef MAPPEDMEMORY
                mem64db[nr]=-1;
            #else
                database[nr]=NULL;
            #endif
            return(false);
        }
    }
#endif*/

    strcat(id,"        ");
    id[8]=0;
    
    printf("loading %s      \r",id);
    fflush(stdout);
    /* make sure that the memory is mapped in ram to prevent excessive swapping */
    
    for(i=0;i<bytesize[nr];i+=256) {
        #ifdef MAPPEDMEMORY
            p=mem64_pointer(mem64db[nr],i);
            *p=0;
        #else
            database[nr][i]=0;
        #endif
    }
    
    if (use_compression==false) {
        #ifdef MAPPEDMEMORY
            mem64_load(mem64db[nr],db_file);
        #else
            fread(database[nr],sizeof(unsigned char),bytesize[nr],db);
        #endif
    }
#ifndef MAPPEDMEMORY
#ifdef USE_ZLIB
    if (use_compression==true) {
        in = gzopen(db_file, "rb");
        if (in == NULL) {
            fprintf(stderr, "can't gzopen %s\n", db_file);
            exit(1);
        }
        len = gzread(in, database[nr], sizeof(unsigned char)*bytesize[nr]);
        if (len!=bytesize[nr]) {
            printf("FATAL: unexpected eof\n");
            exit(1);
        }
        if (gzclose(in) != Z_OK) printf("failed gzclose");
    }
#endif
    fclose(db);
#endif
    if (windows==true) {
        winprint("\n");
        winprint("PROGRESS|*|*|*|*|loading %s\n",id);
    }

    dprint("loading %s      (Size = %4u Mb, total memory = %4u Mb)      \n",id,(int) (bytesize[nr]/1024/1024),(int) (allocatedMemory/1024/1024));
    printf("                                     \r");
    fflush(stdout);
    return(true);
}

int total_db_size(void)
{
    int i;
    int sum=0;

    for(i=0;i<4096;i++) {
        #ifdef MAPPEDMEMORY
            if (mem64db[i]>=0) sum+=sizeof(unsigned char)*bytesize[i];
        #else
            if (database[i]!=NULL) sum+=sizeof(unsigned char)*bytesize[i];
        #endif
    }
    return(sum);
}

void init_databases(void)
{
    int i;
    for(i=0;i<4096;i++) {
        #ifdef MAPPEDMEMORY
            mem64db[i]=-1;
        #else
            database[i]=NULL;
        #endif
        verified[i]=false;
    }
    init_index();
}

void read_all_databases(int maxpieces)
{
    /* reads in all available databases defined in the file 'DB_INDEX_FILE'
    */
    FILE *in;
    char id[12],include;
    int wman,wcrown,bman,bcrown,i,ndb,ii,byte_size,type,n,nr;
    int manc[20]={0,1,  0,1,2,  0,1,2,3,  0,1,2,3,4,   0,1,2,3,4,5};
    int cc[20]=  {1,0,  2,1,0,  3,2,1,0,  4,3,2,1,0,   5,4,3,2,1,0};
    int j;
    int cnt;
    
    in=my_fopen(DB_INDEX_FILE,"r");
    if (in==NULL) {
        printf("error: database index file not found\n");
        return;
    }

    /* read all selected databases */
    for(ii=0;ii<21;ii++) {
        for (j=0;j<21;j++) {
            fscanf(in,"%s",id);
            if (id[0]!='.' && ii>0 && j>0) {
                cnt=manc[ii-1]+cc[ii-1]+manc[j-1]+cc[j-1];
                if (cnt<=maxpieces) {
                    nr=database_nr(white,manc[ii-1],cc[ii-1],manc[j-1],cc[j-1]);
                    #ifdef MAPPEDMEMORY
                    if (mem64db[nr]<0) 
                    #else
                    if (database[nr]==0)
                    #endif
                    {
                        load_database(manc[ii-1],cc[ii-1],manc[j-1],cc[j-1]);
                    }
                }
            }
        }
    }
    fclose(in);
    winprint("\n");
    winprint("PROGRESS|*|*|*|*|\n",id);

}

int database_nr(int color,int wman,int wcrown,int bman,int bcrown)
/* returns the number of the requested database */
{
    int nr;

    if (color==white) nr=wman+8*wcrown+64*bman+512*bcrown;
    else nr=bman+8*bcrown+64*wman+512*wcrown;

    return(nr);
}

int database_retreive_value(int color,int wman,int wcrown,int bman,int bcrown)
/* retreives the value of a database entry using 'database_address' and
   'database_linear_index. 2 bits per position
   
   Returns a value from the perspective of 'color'.
   */
{
    DBINDEX index,dindex;
    int score;
    unsigned char *db;
    static int counter=0;
    int w;
    #ifdef MAPPEDMEMORY
        int handle;
    #endif
    
    /* no pieces: return LOSE */
    if (color==white) if (wman==0 && wcrown==0) return(DB_LOSE);
    if (color==black) if (bman==0 && bcrown==0) return(DB_LOSE);
    #ifdef MAPPEDMEMORY
        handle=mem64db[database_nr(color,wman,wcrown,bman,bcrown)];
        #ifdef PARTIAL
            if (handle<0) return(DB_UNKNOWN);
        #endif
        index=database_linear_index(color);
        dindex=index/4;
        if (dindex<0) {
            printf("fatal: exception 1, %llu\n",dindex);
            exit(1);
        }
        db=mem64_pointer(handle,dindex);

        switch(index-4*dindex) {
            case 0:{score=(*db >>6)&3;break;}
            case 1:{score=(*db >>4)&3;break;}
            case 2:{score=(*db >>2)&3;break;}
            case 3:{score=(*db) &3;break;}
        }
        
    #else
        db=database[database_nr(color,wman,wcrown,bman,bcrown)];
        #ifdef PARTIAL
            if (db==NULL) return(DB_UNKNOWN);
        #endif
        index=database_linear_index(color);
        /*{
            perc=60.0*index/db_count(wman,wcrown,bman,bcrown);
            for (w=0;w<61;w++) {
                if (w<perc || w>(perc+1.0)) printf(".");
                else printf("x");
            }
            printf("%i%i%i%i%i\n",color,wman,wcrown,bman,bcrown);
            display_board();
            counter++;
            
        }*/
        /* place decompression here ! */
        dindex=index/4;
    
        switch(index-4*dindex) {
            case 0:{score=(db[dindex] >>6)&3;break;}
            case 1:{score=(db[dindex] >>4)&3;break;}
            case 2:{score=(db[dindex] >>2)&3;break;}
            case 3:{score=(db[dindex]) &3;break;}
        }
    #endif
    
    return(score);
}

#ifdef MAPPEDMEMORY
    int read_value(int handle,DBINDEX index)
    {
        DBINDEX dindex;
        int score;
        dindex=index/4;
        unsigned char *p;
        
        if (dindex<0) {
            printf("fatal: exception 2, %llu\n",dindex);
             display_board();
            exit(1);
        }

        p=mem64_pointer(handle,dindex);
        
        switch(index-4*dindex) {
            case 0:{score=(*p >>6)&3;break;}
            case 1:{score=(*p >>4)&3;break;}
            case 2:{score=(*p >>2)&3;break;}
            case 3:{score=*p &3;break;}
        }
        return(score);
    }
#else
    int read_value(unsigned char *db,DBINDEX index)
    {
        DBINDEX dindex;
        int score;
        dindex=index/4;
    
        switch(index-4*dindex) {
        case 0:{score=(db[dindex] >>6)&3;break;}
        case 1:{score=(db[dindex] >>4)&3;break;}
        case 2:{score=(db[dindex] >>2)&3;break;}
        case 3:{score=(db[dindex]) &3;break;}
        }
        return(score);
    }
#endif

#ifdef MAPPEDMEMORY
    void store_value(int handle,DBINDEX index,int score)
    {
        DBINDEX dindex;
        unsigned char *p;
        
        dindex=index/4;
        if (dindex<0) {
            printf("fatal: exception 3, %llu\n",dindex);
            exit(1);
        }

        p=mem64_pointer(handle,dindex);
        
        switch(index-4*dindex) {
            case 0:{
                    *p=((*p)&63)+(score*64);
                    break;
                }
            case 1:{
                    *p=((*p)&207)+(score*16);
                    break;
                }
            case 2:{
                    *p=((*p)&243)+(score*4);
                    break;
                }
            case 3:{
                    *p=((*p)&252)+(score);
                    break;
                }
            }
    }
#else
    void store_value(unsigned char *db,DBINDEX index,int score)
    {
        DBINDEX dindex;
        dindex=index/4;
    
        switch(index-4*dindex) {
        case 0:{
                db[dindex]=(db[dindex]&63)+(score*64);
                break;
            }
        case 1:{
                db[dindex]=(db[dindex]&207)+(score*16);
                break;
            }
        case 2:{
                db[dindex]=(db[dindex]&243)+(score*4);
                break;
            }
        case 3:{
                db[dindex]=(db[dindex]&252)+(score);
                break;
            }
        }
    }
}
#endif

void free_all(int skip1,int skip2)
{
    int nr;
    char *id;
    int wman,wcrown,bman,bcrown;
    int freeThis;
    
    for (nr=0;nr<4096;nr++) {
        freeThis=false;
        #ifdef MAPPEDMEMORY
            if (mem64db[nr]>=0 && nr!=skip1 && nr!=skip2) freeThis=true;
        #else
            if (database[nr]!=NULL && nr!=skip1 && nr!=skip2) freeThis=true;
        #endif
        
        if (freeThis==true) {
            allocatedMemory-=sizeof(unsigned char)*bytesize[nr];
            wman=nr & 7;
            wcrown=(nr/8) & 7;
            bman=(nr/64) &7;
            bcrown=(nr/512) & 7;
            id=database_name(wman,wcrown,bman,bcrown);
            strcat(id,"        ");
            id[8]=0;
            printf("freeing %s      (Size = %4i Mb, total memory = %4u Mb)  \n",id,-(int)bytesize[nr]/1024/1024,allocatedMemory/1024/1024);
            #ifdef MAPPEDMEMORY
                mem64_free(mem64db[nr]);
                mem64db[nr]=-1;
            #else
                free(database[nr]);
                database[nr]=NULL;
            #endif
        }
    }
}

void makeSureLoaded(int wman,int wcrown,int bman,int bcrown)
/* makes sure that the given database is loaded into memory */
{
   int nr;
   nr=database_nr(white,wman,wcrown,bman,bcrown);
   #ifdef MAPPEDMEMORY
       if (mem64db[nr]<0) {
           load_database(wman,wcrown,bman,bcrown);
       }
   #else
       if (database[nr]==0) {
           load_database(wman,wcrown,bman,bcrown);
       }
   #endif
}

void countvalue(int wman,int wcrown,int bman,int bcrown)
/* prints all position in a database with value 'value'. */
{
    int nr;
    DBINDEX size;
    DBINDEX index;
    int piecelist[20];
    DBINDEX win=0,draw=0,lose=0,unknown=0;

    nr=database_nr(white,wman,wcrown,bman,bcrown);
    size=pos_count[nr];

    for(index=0;index<size;index++) {
        #ifdef MAPPEDMEMORY
            if (read_value(mem64db[nr],index)==DB_WIN) win++;
            if (read_value(mem64db[nr],index)==DB_DRAW) draw++;
            if (read_value(mem64db[nr],index)==DB_LOSE) lose++;
            if (read_value(mem64db[nr],index)==DB_UNKNOWN) unknown++;
        #else
            if (read_value(database[nr],index)==DB_WIN) win++;
            if (read_value(database[nr],index)==DB_DRAW) draw++;
            if (read_value(database[nr],index)==DB_LOSE) lose++;
            if (read_value(database[nr],index)==DB_UNKNOWN) unknown++;
        #endif
    }
    printf("win %9s  ",neatNumber(win));
    printf("draw %9s  ",neatNumber(draw));
    printf("lose %9s  ",neatNumber(lose));
    printf("unknown/illegal %s\n",neatNumber(unknown));
}

int nextboard(int *list,int *type,int *constraint,int *min,int *max,int p)
{
    int try,stop=false;

    if (p<0) return(true);
    board[map[list[p]]]=empty;
    do {
        list[p]++;
    } while(board[map[list[p]]]!=empty && list[p]<max[p]);
    if (list[p]>=max[p]) {
back:
        stop=nextboard(list,type,constraint,min,max,p-1);
        try=min[p];
        if (constraint[p]==1) try=list[p-1]+1;
        while(board[map[try]]!=empty && try<max[p]) try++;
        if (try>=max[p]) goto back;
        list[p]=try;
    }
    board[map[list[p]]]=type[p];
    return(stop);
}

int init_nextboard(int *list,int *type,int *constraint,int *min,int *max,int wman,int wcrown,int bman,int bcrown)
{
    int i,j,n;

    init_board();
    for(i=0;i<50;i++) board[map[i]]=empty;
    for(i=0;i<12;i++) list[i]=49;
    for(i=0;i<12;i++) constraint[i]=1;
    j=0;
    constraint[j]=0;
    for(i=0;i<bman;i++) {min[j]=0; max[j]=45; type[j++]=black|man;}
    constraint[j]=0;
    for(i=0;i<wman;i++) {min[j]=5; max[j]=50; type[j++]=white|man;}
    constraint[j]=0;
    for(i=0;i<bcrown;i++) {min[j]=0; max[j]=50; type[j++]=black|crown;}
    constraint[j]=0;
    for(i=0;i<wcrown;i++) {min[j]=0; max[j]=50; type[j++]=white|crown;}
    n=j;
    nextboard(list,type,constraint,min,max,n-1);
    return(n);
}

int determen_value(int nr,DBINDEX index)
/* determens and sets, if possible, a value in database 'nr',
   returns the value
*/
{
    int score,nmoves,m,found=DB_UNKNOWN,known,best;

    best=DB_LOSE; known=true;
    nmoves=move_list(0,white);
    for(m=0;m<nmoves;m++) {
        do_move(movelist[0][m]);
        score=database_retreive_value(black,pieces[white|man],pieces[white|crown],pieces[black|man],pieces[black|crown]);
        undo_move(movelist[0][m]);
        if (score==DB_UNKNOWN) known=false;
    else if (score==DB_LOSE) {best=DB_WIN; break;}
        else if (score==DB_DRAW) best=DB_DRAW;
    }
    if (best==DB_WIN || known==true) {
        #ifdef MAPPEDMEMORY
            store_value(mem64db[nr],index,best);
        #else
            store_value(database[nr],index,best);
        #endif
        found=best;
    }
    return(found);
}

DBINDEX one_iteration_fromqueue(int nr,int wman,int wcrown,int bman,int bcrown,int inq,int outq,int mirror)
{
    int nmoves,m,result;
    DBINDEX index,found=0;
    DBINDEX i,j;
    int starttime;
    static char *local;

    starttime=clock();
    init_queue_write(outq);
    init_queue_read(inq);
    init_board();

    for(i=0;i<qsize[inq];i++) {
        if ((i&1023)==0) {
            printf("backtracking:  %0.4f  (da: %llu Mb)         \r",(float)i/qsize[inq],mem64_diskActivity/1024/1024);
            fflush(stdout);
            }
        local=read_from_queue(inq);
        /* we now have a new board position, with black to move. Undo the
           last (white) move and do a forward search. If results, store into
           the out-queue
           */
        if (mirror==0) for(j=0;j<50;j++) board[map[j]]=local[map[j]];
        else reverse_board(board,local);

        set_pieces();

        nmoves=reverse_move_list(1,white); /* kings only ! */

        for(m=0;m<nmoves;m++) {
            do_move(movelist[1][m]); /* we have a potential position, try it ! */
            set_pieces();
            index=database_linear_index(white);
            #ifdef MAPPEDMEMORY
                if (read_value(mem64db[nr],index)==DB_UNKNOWN)
            #else
                if (read_value(database[nr],index)==DB_UNKNOWN)
            #endif            
            
            {
                result=determen_value(nr,index);
#ifdef PARTIAL
                if (result==DB_WIN || result==DB_LOSE || result==DB_DRAW) found++;
                if (result==DB_DRAW) add_to_queue(outq);
#else
                if (result==DB_WIN || result==DB_LOSE) found++;
#endif
                if (result==DB_WIN || result==DB_LOSE) add_to_queue(outq);
            }
            undo_move(movelist[1][m]);
        }
    }
    close_queue(inq);
    close_queue(outq);
    printf("\007backward: found=%s, time=%i                      \n",neatNumber(found),(clock()-starttime)/CLOCKS_PER_SEC); fflush(stdout);
    fprintf(logfile,"db:%i%i%i%i, found=%i, time=%i\n",wman,wcrown,bman,bcrown,found,(clock()-starttime)/CLOCKS_PER_SEC); fflush(logfile);

    return(found);
}

DBINDEX one_iteration(int nr,int wman,int wcrown,int bman,int bcrown,int mode,int outq)
/* does one iteration in creating database 'nr'. Returns the number of
   newly discovered positions */
{
    DBINDEX index,i,j,n,size,found=0;
    int score,trypos;
    DBINDEX ni=0;
    DBINDEX total;
    int piecelist[40]; /* type of piece for 1 ... n */
    int list[12],type[12],constraint[12],min[12],max[12];
    int stop=false,starttime,result;
    FILE *out=NULL;

    starttime=clock();
    total=db_count(wman,wcrown,bman,bcrown);

    if (outq>=0) init_queue_write(outq);
    n=init_nextboard(list,type,constraint,min,max,wman,wcrown,bman,bcrown);
    size=pos_count[nr];
    database_piecelist(piecelist,wman,wcrown,bman,bcrown);
    set_pieces();
    do {
        index=database_linear_index(white);
        if (index<0) { printf("fatal index: %llu\n",index); display_board();}
        trypos=true;
        #ifdef MAPPEDMEMORY
            if (read_value(mem64db[nr],index)!=DB_UNKNOWN) trypos=false;
        #else
            if (read_value(database[nr],index)!=DB_UNKNOWN) trypos=false;
        #endif
        else {
            if (mode==CAPTURE) { /* capture moves only */
                set_pieces();
                /*printf("%i\n",ni); display_board();fflush(stdout);*/
                if (quiet(white)==true) trypos=false;
            }
            if (mode==PROMOTE) { /* promotion moves only */
                set_pieces();
                /*printf("%i\n",ni); display_board();fflush(stdout);*/
                if (!has_promote(white)) trypos=false;
            }
        }
        if (trypos==true) {
            /*if (mode==PROMOTE) {printf("%i\n",ni); display_board();fflush(stdout);}*/
            result=determen_value(nr,index);
#ifdef PARTIAL
            if (result==DB_WIN || result==DB_LOSE || result==DB_DRAW) found++;
            if (result==DB_DRAW && outq>=0) add_to_queue(outq);
#else
            if (result==DB_WIN || result==DB_LOSE) found++;
#endif
            if (result==DB_WIN || result==DB_LOSE && outq>=0) add_to_queue(outq);
        }
        ni++;
    if ((ni&1023)==0) {
        if (mode==CAPTURE) printf("captures only: %0.4f  (da: %llu Mb)   \r",(float)ni/total,mem64_diskActivity/1024/1024);
        else if (mode==PROMOTE) printf("promotions:    %0.4f  (da: %llu Mb)   \r",(float)ni/total,mem64_diskActivity/1024/1024);
        else printf("all moves:     %0.4f  (da: %llu Mb)   \r",(float)ni/total,mem64_diskActivity/1024/1024);
        fflush(stdout);            
        }
    } while(nextboard(list,type,constraint,min,max,n-1)==false);
    printf("\007forward: found=%s, time=%i              \n",neatNumber(found),(clock()-starttime)/CLOCKS_PER_SEC);
    fflush(stdout);
    fprintf(logfile,"db:%i%i%i%i, found=%i, time=%i\n",wman,wcrown,bman,bcrown,found,(int)(clock()-starttime)/CLOCKS_PER_SEC); fflush(logfile);
    if (outq>=0) close_queue(outq);
    return(found);
}

void validqueue(int *q1,int *q2)
{
    /* if all went ok, q2 contains a valid queue: use this as
       input queue
       */
    *q1=1-*q1; *q2=1-*q2;
}

int availableOnDisk(int wman,int wcrown,int bman,int bcrown)
/* returns true if database is available on disk */
{
    char db_filename[100],*id;
    FILE *in;

    id=database_name(wman,wcrown,bman,bcrown);
    #ifdef USE_ZLIB
        sprintf(db_filename,"databases/%s.raw.gz",id);
    #else
        sprintf(db_filename,"databases/%s.raw",id);
    #endif
    in=fopen(db_filename,"rb");
    if (in==NULL) return (false);
    return (true);
}



void create_database(int wman,int wcrown,int bman,int bcrown,int level)
/*
    Provides the general structure of the solving program; it makes sure
    that the order of database generation is correct and takes care of the
    memory management and loading/storing databases.
*/
{
    int wm,wc,bm,bc;
    unsigned char *db;
    char db_file[2][100],*id;
    int nr[2];
    DBINDEX size[2];
    DBINDEX p,n,i;
    DBINDEX pos,found=0,newfound,loaded=0;
    int q1,q2;
    float startingTime;
    int onDisk;

    FILE *out,*in;
    int mirror;  /* mirror is set to 1 for symmetric databases, 2 otherwise */

    if (level<0) return;
    
    /* has it been created ? */
    nr[0]=database_nr(white,wman,wcrown,bman,bcrown);
    nr[1]=database_nr(white,bman,bcrown,wman,wcrown);

    /* filenames */
    id=database_name(wman,wcrown,bman,bcrown);
    #ifdef USE_ZLIB
        sprintf(db_file[0],"databases/%s.raw.gz",id);
    #else
        sprintf(db_file[0],"databases/%s.raw",id);
    #endif
    id=database_name(bman,bcrown,wman,wcrown);
    #ifdef USE_ZLIB
        sprintf(db_file[1],"databases/%s.raw.gz",id);
    #else
        sprintf(db_file[1],"databases/%s.raw",id);
    #endif

    /* determen size of the database(s) */
    size[0]=db_count(wman,wcrown,bman,bcrown);
    size[1]=db_count(bman,bcrown,wman,wcrown);

    /* symmetric database ? */
    if (nr[0]==nr[1]) mirror=1; else mirror=2;

    /* check if database is already loaded */
    #ifdef MAPPEDMEMORY
        if (mem64db[nr[0]]>=0 && mem64db[nr[1]]>0) { /* database is already present */
        return;
    }
    #else
        if (database[nr[0]]!=NULL && database[nr[1]]!=NULL) { /* database is already present */
            return;
        }
    #endif
    /* if possible: load database from disk */
    if (availableOnDisk(wman,wcrown,bman,bcrown)) {
        load_database(wman,wcrown,bman,bcrown);
        return;
    }

    /* current database is not loaded, nor on disk, so we will be creating the
       database now.
       
       First we have to make sure that we created all required sub-databases.
       Dont load any from disk, we will do that on demand later.
    */

    /* database 0: capture 1 or more black pieces */
    for(bm=0;bm<=bman;bm++)
        for (bc=0;bc<=bcrown;bc++)
            if ((bc > 0 || bm > 0) && !(bm == bman && bc == bcrown)) {
                if (!availableOnDisk(bm,bc,wman,wcrown)) create_database(bm,bc,wman,wcrown,level-1);
                if (wman>0 && !availableOnDisk(bm,bc,wman-1,wcrown+1)) create_database(bm,bc,wman-1,wcrown+1,level-1);  /* capture AND promote */
            }
    /* database 0: promote a white man */
    if (wman>0 && !availableOnDisk(bman,bcrown,wman-1,wcrown+1)) create_database(bman,bcrown,wman-1,wcrown+1,level-1);

    /* database 1: capture 1 or more white pieces */
    for(wm=0;wm<=wman;wm++)
        for (wc=0;wc<=wcrown;wc++)
            if ((wc > 0 || wm > 0) && !(wm == wman && wc == wcrown)) {
                if (!availableOnDisk(wm,wc,bman,bcrown)) create_database(wm,wc,bman,bcrown,level-1);
                if (bman>0 && !availableOnDisk(wm,wc,bman-1,bcrown+1)) create_database(wm,wc,bman-1,bcrown+1,level-1);
            }

    /* promote a black man */
    if (bman>0 && !availableOnDisk(wman,wcrown,bman-1,bcrown+1)) create_database(wman,wcrown,bman-1,bcrown+1,level-1);

    /* free all memory */
    free_all(-1,-1);
    
    /* initialise queue system */
    q1=0; q2=1;
    qmax[0]=0.1*size[0];
    if (size[0]>mem64_RAM()) {
        /* if database doesn't fit memory, we are less keen on doiing a backward search because
           it runs trough the ram more randomly than a forward search */
        qmax[0]=0.02*size[0];
    }    
    if (qmax[0]>QMAX) qmax[0]=QMAX;
    qmax[1]=qmax[0];
    qsize[0]=qsize[1]=-1LL;
    startingTime=clock();
    found=0;
    
    if (mirror==1) {
        /* first pass: capture moves only */
        
        pos_count[nr[0]]=size[0];
        bytesize[nr[0]]=(size[0]+3)/4;
        #ifdef MAPPEDMEMORY
            mem64db[nr[0]]=mem64_allocate(bytesize[nr[0]]);
        #else
            database[nr[0]]=(unsigned char *)malloc(sizeof(unsigned char)*bytesize[nr[0]]);
        #endif
        allocatedMemory+=sizeof(unsigned char)*bytesize[nr[0]];

        set_col(33,33);
        printf("creating %s, %s positions (Total memory alloc: %u Mb)   \n",database_name(wman,wcrown,bman,bcrown),neatNumber(size[0]),allocatedMemory/1024/1024);
        res_col();

        #ifdef MAPPEDMEMORY
            if (mem64db[nr[0]]<0)  {
                printf("fatal: no memory for database (%i Mb)\n",bytesize[nr[0]]/1024/1024);
                exit(1);
            }
        #else
            if (database[nr[0]]==NULL) {
                printf("fatal: no memory for database (%i Mb)\n",bytesize[nr[0]]/1024/1024);
                exit(1);
            }
        #endif
        
        for(p=0;p<size[1];p++) {
            #ifdef MAPPEDMEMORY
                store_value(mem64db[nr[1]],p,DB_UNKNOWN);
            #else
                store_value(database[nr[1]],p,DB_UNKNOWN);
            #endif
        }

        for(bm=0;bm<=bman;bm++)
            for (bc=0;bc<=bcrown;bc++)
                if ((bc > 0 || bm > 0) && !(bm == bman && bc == bcrown)) {
                    load_database(bm,bc,wman,wcrown);
                    load_database(wman,wcrown,bm,bc);
                    if (wman>0) {
                        load_database(bm,bc,wman-1,wcrown+1);
                        load_database(wman-1,wcrown+1,bm,bc);
                    }
                }
        if (wman>0) {
            load_database(bman,bcrown,wman-1,wcrown+1);
            load_database(wman-1,wcrown+1,bman,bcrown);
        }
        found=one_iteration(nr[0],wman,wcrown,bman,bcrown,CAPTURE,-1);

        /* second pass: force all positions being seen */
                
        found+=newfound=one_iteration(nr[0],wman,wcrown,bman,bcrown,NORMAL,q2);
        validqueue(&q1,&q2);

        while(newfound>0) {
            if (qsize[q1]<=0) {
                newfound=one_iteration(nr[0],wman,wcrown,bman,bcrown,NORMAL,q2); /* no queue available */\
                found+=newfound;
            }
            else {
                found+=newfound=one_iteration_fromqueue(nr[0],wman,wcrown,bman,bcrown,q1,q2,mirror); /* previous computation resulted in a queue */
            }
            validqueue(&q1,&q2);
        }
    }
    else {
        /* initialise white to move database */
        pos_count[nr[0]]=size[0];
        bytesize[nr[0]]=(size[0]+3)/4;
        #ifdef MAPPEDMEMORY
            mem64db[nr[0]]=mem64_allocate(bytesize[nr[0]]);
        #else
            database[nr[0]]=(unsigned char *)malloc(sizeof(unsigned char)*bytesize[nr[0]]);
        #endif
        allocatedMemory+=sizeof(unsigned char)*bytesize[nr[0]];
        set_col(33,33);
        printf("creating %s, %s positions (Total memory alloc: %u Mb)   \n",database_name(wman,wcrown,bman,bcrown),neatNumber(size[0]),allocatedMemory/1024/1024);
        res_col();
        #ifdef MAPPEDMEMORY
            if (mem64db[nr[0]]<0) {
                printf("fatal: no memory for database (%i Mb)\n",bytesize[nr[0]]/1024/1024);
                exit(1);
            }
        #else
            if (database[nr[0]]==NULL) {
                printf("fatal: no memory for database (%i Mb)\n",bytesize[nr[0]]/1024/1024);
                exit(1);
            }
        #endif
        
        for(p=0;p<size[0];p++) {
            #ifdef MAPPEDMEMORY
                store_value(mem64db[nr[0]],p,DB_UNKNOWN);
            #else
                store_value(database[nr[0]],p,DB_UNKNOWN);
            #endif
        }

        /* first pass: capture moves only: white to move*/
        for(bm=0;bm<=bman;bm++)
            for (bc=0;bc<=bcrown;bc++)
                if ((bc > 0 || bm > 0) && !(bm == bman && bc == bcrown)) {
                    load_database(bm,bc,wman,wcrown);
                    if (wman>0) load_database(bm,bc,wman-1,wcrown+1);
                }
        found+=one_iteration(nr[0],wman,wcrown,bman,bcrown,CAPTURE,-1);
        free_all(nr[0],nr[1]);
        
        /* initialise black to move database */
        pos_count[nr[1]]=size[1];
        bytesize[nr[1]]=(size[1]+3)/4;
        #ifdef MAPPEDMEMORY
            mem64db[nr[1]]=mem64_allocate(bytesize[nr[1]]);
        #else
            database[nr[1]]=(unsigned char *)malloc(sizeof(unsigned char)*bytesize[nr[1]]);
        #endif
        allocatedMemory+=sizeof(unsigned char)*bytesize[nr[1]];
        set_col(33,33);
        printf("creating %s, %s positions (Total memory alloc: %i Mb)   \n",database_name(bman,bcrown,wman,wcrown),neatNumber(size[1]),allocatedMemory/1024/1024);
        res_col();
        #ifdef MAPPEDMEMORY
            if (mem64db[nr[1]]<0) {
                printf("fatal: no memory for database (%i Mb)\n",bytesize[nr[1]]/1024/1024);
                exit(1);
            }
        #else
            if (database[nr[1]]==NULL) {
                printf("fatal: no memory for database (%i Mb)\n",bytesize[nr[1]]/1024/1024);
                exit(1);
            }
        #endif
        for(p=0;p<size[1];p++) {
            #ifdef MAPPEDMEMORY
                store_value(mem64db[nr[1]],p,DB_UNKNOWN);
            #else
                store_value(database[nr[1]],p,DB_UNKNOWN);
            #endif
        }

        /* first pass: capture moves only: black to move */
        for(wm=0;wm<=wman;wm++)
            for (wc=0;wc<=wcrown;wc++)
                if ((wc > 0 || wm > 0) && !(wm == wman && wc == wcrown)) {
                    load_database(wm,wc,bman,bcrown);
                    if (bman>0) load_database(wm,wc,bman-1,bcrown+1);
                }
        found+=one_iteration(nr[1],bman,bcrown,wman,wcrown,CAPTURE,-1);
        free_all(nr[0],nr[1]);
        
        /* load databases for moves with promotions but no captures */
        
        if (wman>0) {
            load_database(bman,bcrown,wman-1,wcrown+1);
        }
        if (bman>0) {
            load_database(wman,wcrown,bman-1,bcrown+1);
        }

        /* second pass: force all positions being seen */
        
        found+=one_iteration(nr[0],wman,wcrown,bman,bcrown,NORMAL,-1);
        
        found+=one_iteration(nr[1],bman,bcrown,wman,wcrown,NORMAL,q2);
        validqueue(&q1,&q2);

        do {
            if (qsize[q1]<=0) found+=newfound=one_iteration(nr[0],wman,wcrown,bman,bcrown,NORMAL,q2);
            else found+=newfound=one_iteration_fromqueue(nr[0],wman,wcrown,bman,bcrown,q1,q2,mirror);
            validqueue(&q1,&q2);
            if (newfound==0) break;
            if (qsize[q1]<=0) found+=newfound=one_iteration(nr[1],bman,bcrown,wman,wcrown,NORMAL,q2);
            else found+=newfound=one_iteration_fromqueue(nr[1],bman,bcrown,wman,wcrown,q1,q2,mirror);
            validqueue(&q1,&q2);
        } while(newfound!=0);
    }
    /* wrapping up */
   
    printf("Total time: time=%.1f\n",(clock()-startingTime)/CLOCKS_PER_SEC);
    fprintf(logfile,"Total time: =%.1f\n",(clock()-startingTime)/CLOCKS_PER_SEC);
    printf("a total of %i positions have been solved\n",found);
    printf("Accumulative disk activity:%llu Mb\n",mem64_diskActivity/1024/1024);
    
    fflush(stdout);
    countvalue(wman,wcrown,bman,bcrown);
    if (mirror==2) countvalue(bman,bcrown,wman,wcrown);

#ifndef PARTIAL
    for(i=0;i<mirror;i++) for(p=0;p<size[i];p++) {
        #ifdef MAPPEDMEMORY
            if (read_value(mem64db[nr[i]],p)==DB_UNKNOWN)
                store_value(mem64db[nr[i]],p,DB_DRAW);
        #else
            if (read_value(database[nr[i]],p)==DB_UNKNOWN)
                store_value(database[nr[i]],p,DB_DRAW);
        #endif
        }
#endif

    /* write databases */
    for(i=0;i<mirror;i++) {
        printf("writing file '%s'\n",db_file[i]);
        #ifdef MAPPEDMEMORY
            mem64_save(mem64db[nr[i]],db_file[i]);
        #else
            out=fopen(db_file[i],"wb");
            if (out==NULL) {
                printf("fatal: write error\n");
                exit(1);
            }
            fwrite(database[nr[i]],sizeof(unsigned char),bytesize[nr[i]],out);
            fclose(out);
        #endif
    }
}

void debug(int value,int wman,int wcrown,int bman,int bcrown)
/* prints all position in a database with value 'value'. */
{
    DBINDEX nr,size,i,j,n;
    DBINDEX index,count;
    int list[12],type[12],constraint[12],min[12],max[12];
    char *id;
    int best,known,original,score,nmoves,m;
    DBINDEX total;
    count=0;

    id=database_name(wman,wcrown,bman,bcrown);
    total=db_count(wman,wcrown,bman,bcrown);
    set_col(33,33);
    printf("debug %s, searching value %i\n",id,value);
    fprintf(logfile,"debug %s, searching value %i\n",id,value);
    res_col();
    nr=database_nr(white,wman,wcrown,bman,bcrown);
    #ifdef MAPPEDMEMORY
        if (mem64db[nr]<0)  {
            printf("database is not present\n");
            return;
        }
    #else
        if (database[nr]==0) {
            printf("database is not present\n");
            return;
        }
    #endif
    
    countvalue(wman,wcrown,bman,bcrown);
    n=init_nextboard(list,type,constraint,min,max,wman,wcrown,bman,bcrown);
    i=0;
    do {
        index=database_linear_index(white);
        if (value==DB_VERIFY) {
            #ifdef MAPPEDMEMORY
                original=read_value(mem64db[nr],index);
            #else
                original=read_value(database[nr],index);
            #endif
            best=DB_LOSE; known=true;
            set_pieces();
            nmoves=move_list(0,white);
            for(m=0;m<nmoves;m++) {
                do_move(movelist[0][m]);
                score=database_retreive_value(black,pieces[white|man],pieces[white|crown],pieces[black|man],pieces[black|crown]);
                undo_move(movelist[0][m]);
                if (score==DB_UNKNOWN) known=false;
            if (score==DB_LOSE) {best=DB_WIN; break;}
                if (score==DB_DRAW) best=DB_DRAW;
            }
            if (known==false || best!=original) {
                printf("error found! nr:%i ply0 value:%i ply1 value:%i, index:%i\n",i,original,best,index);
                display_board();
                i++;
            }
        }
        else
        #ifdef MAPPEDMEMORY
            if (value==DB_ALL || read_value(mem64db[nr],index)==value) {
                display_board();
                printf("index: %i nr:%i  value:%i\n",index,i,read_value(mem64db[nr],index));
                i++;
        }
        #else
            if (value==DB_ALL || read_value(database[nr],index)==value) {
                display_board();
                printf("index: %i nr:%i  value:%i\n",index,i,read_value(database[nr],index));
                i++;
            }
        #endif
        if ((count&1023)==0) {
            printf("%0.4f (da: %llu Mb)           \r",(float)count/total,mem64_diskActivity/1024/1024);
            fflush(stdout);
            }
        count++;
    } while(nextboard(list,type,constraint,min,max,n-1)==false);
    if (value==DB_VERIFY) {
        printf("verified %s errors:%i\n\n",id,i);
        fprintf(logfile,"verified %s errors:%i\n\n",id,i);
    }
}

void verify_all(wman,wcrown,bman,bcrown,level)
{
    int wm,wc,bm,bc,nr[2];

    if (level<0) return;
    nr[0]=database_nr(white,wman,wcrown,bman,bcrown);
    nr[1]=database_nr(white,bman,bcrown,wman,wcrown);
    if (verified[nr[0]]==true) return;
    #ifdef MAPPEDMEMORY
        if (mem64db[nr[0]]<0)
    #else
        if (database[nr[0]]==NULL)
    #endif
        {
        if (availableOnDisk(wman,wcrown,bman,bcrown)) {
            load_database(wman,wcrown,bman,bcrown);
            if (nr[0]!=nr[1]) {
                load_database(bman,bcrown,wman,wcrown);
            }
        }
    }
    /* first verify all sub-databases */
    for(wm=0;wm<=7;wm++) for(wc=0;wc<=7;wc++)
            for(bm=0;bm<=7;bm++) for(bc=0;bc<=7;bc++)
                    if (wm!=wman || bm!=bman || wc!=wcrown || bc!=bcrown)
                        if ((wm+wc)>0 && (bm+bc)>0)
                            if (wm<=wman && bm<=bman && wc<=(wman+wcrown) && bc<=(bman+bcrown))
                                if ((wm+wc)<=(wman+wcrown) && (bm+bc)<=(bman+bcrown))
                                    verify_all(wm,wc,bm,bc,level-1);
    debug(DB_VERIFY,wman,wcrown,bman,bcrown);
    if (nr[0]!=nr[1]) debug(DB_VERIFY,bman,bcrown,wman,wcrown);
    verified[nr[0]]=verified[nr[1]]=true;
}

int database_value(int color,int wman,int wcrown,int bman,int bcrown)
/* checks of board[] is in a loaded database. If it is, it returns
   one of:WIN,LOSE or UNKNOWN, seen from the white party. Inputs:
   color: player to move
   wman,wcrown,bman,bcrown: the number of these pieces on the board.
   In general these are known, so it speeds up the access.
*/
{
    DBINDEX index;
    int score,nr;
    unsigned char *db;

    if (use_db==false) return(UNKNOWN);
    nr=database_nr(color,wman,wcrown,bman,bcrown);
    #ifdef MAPPEDMEMORY
        if (mem64db[nr]<0) return(UNKNOWN);
    #else
        db=database[nr];
        if (db==NULL) return(UNKNOWN);
    #endif
    
    index=database_linear_index(color);
    #ifdef MAPPEDMEMORY
        score=read_value(mem64db[nr],index);
    #else
        score=read_value(db,index);
    #endif
    
    if (score==DB_UNKNOWN) {dbfail++; return(UNKNOWN);}
    ndat++; db_usage[nr]++;
    if (color==white) {
        if (score==DB_DRAW) return(DRAW);
        if (score==DB_WIN) return(WIN);
        if (score==DB_LOSE) return(LOSE);
    }
    else {
        if (score==DB_DRAW) return(DRAW);
        if (score==DB_WIN) return(LOSE);
        if (score==DB_LOSE) return(WIN);
    }
    return(UNKNOWN); /* never get here */
}

int db_main()
{
    char input[1024];
    float startTime;
    DBINDEX test;
    
    logfile=fopen("log.txt","a");
    if (logfile==NULL) {
        printf("error 23\n");
        exit(1);
    }
    
    printf("Dragon database generator.\n");
    printf("Type 'help' for help.\n\n");
    printf("\nIndex size:%i bit\n",8*sizeof(DBINDEX));
    init_var();
    init_databases();
    mem64_init(true);
    /*mem64_test();*/
    
    do {
        int wman,wcrown,bman,bcrown;

        printf("generate> ");
        scanf("%s",input);
        if (input[0]=='!') system(&input[1]);
        else if (strcmp(input,"help")==0) {
            printf("\
                   help                                        generate this mesage\n\
                   make {#white man} {wcrown} {bman} {bcrown}  create database\n\
                   find {value} {wm wc} {bm} {bc}              find positions with\n\
                                                               given value: 0=lose,\n\
                                                               1=win,2=lose.\n\
                   verify {wm wc} {bm} {bc}                    Verify consistency of\n\
                                                               given and sub-databases.\n\
                   count {wm wc} {bm} {bc}                     NUmber of positions in database\n\
                   exit                                        Quit program\n\
                   load {wm} {wc} {bm} {bc}                    Load database into memory\n\
                   freeall                                     Free all memory\n\
                   countval {wm} {wc} {bm} {bc}                Show number of wins/draw/loses in database\n\
                   dumpq {queue 0/1} {number}                  Show the positions in the queue\n\
                   \n");
        }
        else if (strcmp(input,"make")==0) {
            fprintf(logfile,"Starting\n");
            scanf("%i%i%i%i",&wman,&wcrown,&bman,&bcrown);
            startTime=clock();
            fprintf(logfile,"Starting\n");
            startTime=clock();
            create_database(wman,wcrown,bman,bcrown,10);
            fprintf(logfile,"Total time: %.1f\n",(clock()-startTime)/CLOCKS_PER_SEC);
        }
        else if (strcmp(input,"find")==0) {
            int value,nr;
            scanf("%i%i%i%i%i",&value,&wman,&wcrown,&bman,&bcrown);
            makeSureLoaded(wman,wcrown,bman,bcrown);
            debug(value,wman,wcrown,bman,bcrown);
        }
        else if (strcmp(input,"count")==0) {
            scanf("%i%i%i%i",&wman,&wcrown,&bman,&bcrown);
            test=db_count(wman,wcrown,bman,bcrown);
            printf("%s\n",neatNumber(test));
        }
        else if (strcmp(input,"load")==0) {
            scanf("%i%i%i%i",&wman,&wcrown,&bman,&bcrown);
            load_database(wman,wcrown,bman,bcrown);
        }
        else if (strcmp(input,"countval")==0) {
            scanf("%i%i%i%i",&wman,&wcrown,&bman,&bcrown);
            makeSureLoaded(wman,wcrown,bman,bcrown);
            countvalue(wman,wcrown,bman,bcrown);
        } 
        else if (strcmp(input,"dumpq")==0) {
            int q;
            static char *local;
            int i,j,n;

            scanf("%i%i",&q,&n);
            qsize[q]=n;
            init_queue_read(q);
            for (i=0;i<n;i++) {
                printf("%i\n",i);
                local=read_from_queue(q);
                for(j=0;j<50;j++) board[map[j]]=local[map[j]];
                display_board();
            }
        }
        else if (strcmp(input,"fdumpq")==0) {
            int q;
            static char *local;
            int i,j,n;
            char name[100];
            
            scanf("%i%i",&q,&n);
            qsize[q]=n;
            init_queue_read(q);
            for (i=0;i<n;i++) {
                local=read_from_queue(q);
                for(j=0;j<50;j++) board[map[j]]=local[map[j]];
                sprintf(name,"queue-%i.dcp",i);
                printf("%s\n",name);
                save_board(white,name);
            }
        }
        else if (strcmp(input,"freeall")==0) {
            free_all(-1,-1);
        }
        else if (strcmp(input,"verify")==0) {
            int value,nr;
            scanf("%i%i%i%i",&wman,&wcrown,&bman,&bcrown);
            startTime=clock();
            fprintf(logfile,"Starting\n");
            verify_all(wman,wcrown,bman,bcrown,10);
            fprintf(logfile,"Total time: %.1f\n",(clock()-startTime)/CLOCKS_PER_SEC);
        }

    } while(strcmp(input,"quit")!=0 && strcmp(input,"q")!=0 && strcmp(input,"exit")!=0);
    /*debug  (DB_LOSE,3,1,1,0);

    verify_all(3,0,3,0,10);
    verify_all(4,0,1,0,10);*/


    /* create_database(4,0,1,0,0);

     debug  (DB_LOSE,4,0,1,0);*/
    fclose(logfile);
    #ifdef MAPPEDMEMORY
        mem64_exit();
    #endif
    return(0);
}
